# """
#     payload.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import json
import base64
from typing import Any, Optional
from snowplow_tracker.typing import PayloadDict, JsonEncoderFunction


class Payload:
    def __init__(self, dict_: Optional[PayloadDict] = None) -> None:
        """
        Constructor
        """

        self.nv_pairs = {}

        if dict_ is not None:
            for f in dict_:
                self.nv_pairs[f] = dict_[f]

    """
    Methods to add to the payload
    """

    def add(self, name: str, value: Any) -> None:
        """
        Add a name value pair to the Payload object
        """
        if not (value == "" or value is None):
            self.nv_pairs[name] = value

    def add_dict(self, dict_: PayloadDict, base64: bool = False) -> None:
        """
        Add a dict of name value pairs to the Payload object

        :param  dict_:          Dictionary to be added to the Payload
        :type   dict_:          dict(string:\\*)
        """
        for f in dict_:
            self.add(f, dict_[f])

    def add_json(
        self,
        dict_: Optional[PayloadDict],
        encode_base64: bool,
        type_when_encoded: str,
        type_when_not_encoded: str,
        json_encoder: Optional[JsonEncoderFunction] = None,
    ) -> None:
        """
        Add an encoded or unencoded JSON to the payload

        :param  dict_:                  Custom context for the event
        :type   dict_:                  dict(string:\\*) | None
        :param  encode_base64:          If the payload is base64 encoded
        :type   encode_base64:          bool
        :param  type_when_encoded:      Name of the field when encode_base64 is set
        :type   type_when_encoded:      string
        :param  type_when_not_encoded:  Name of the field when encode_base64 is not set
        :type   type_when_not_encoded:  string
        :param json_encoder:            Custom JSON serializer that gets called on non-serializable object
        :type  json_encoder:            function | None
        """

        if dict_ is not None and dict_ != {}:

            json_dict = json.dumps(dict_, ensure_ascii=False, default=json_encoder)

            if encode_base64:
                encoded_dict = base64.urlsafe_b64encode(json_dict.encode("utf-8"))
                if not isinstance(encoded_dict, str):
                    encoded_dict = encoded_dict.decode("utf-8")
                self.add(type_when_encoded, encoded_dict)

            else:
                self.add(type_when_not_encoded, json_dict)

    def get(self) -> PayloadDict:
        """
        Returns the context dictionary from the Payload object
        """
        return self.nv_pairs



# """
#     redis_emitter.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import json
import logging
from typing import Any, Optional
from snowplow_tracker.typing import PayloadDict, RedisProtocol

_REDIS_OPT = True
try:
    import redis
except ImportError:
    _REDIS_OPT = False

# logging
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class RedisEmitter(object):
    """
    Sends Snowplow events to a Redis database
    """

    if _REDIS_OPT:

        def __init__(
            self, rdb: Optional[RedisProtocol] = None, key: str = "snowplow"
        ) -> None:
            """
            :param rdb:  Optional custom Redis database
            :type  rdb:  redis | None
            :param key:  The Redis key for the list of events
            :type  key:  string
            """
            if rdb is None:
                rdb = redis.StrictRedis()

            self.rdb = rdb
            self.key = key

        def input(self, payload: PayloadDict) -> None:
            """
            :param payload:  The event properties
            :type  payload:  dict(string:*)
            """
            logger.debug("Pushing event to Redis queue...")
            self.rdb.rpush(self.key, json.dumps(payload))
            logger.info("Finished sending event to Redis.")

        def flush(self) -> None:
            logger.warning("The RedisEmitter class does not need to be flushed")

        def sync_flush(self) -> None:
            self.flush()

    else:

        def __new__(cls, *args: Any, **kwargs: Any) -> "RedisEmitter":
            logger.error(
                "RedisEmitter is not available. Please install snowplow-tracker with redis extra dependency."
            )
            raise RuntimeError(
                "RedisEmitter is not available. To use: `pip install snowplow-tracker[redis]`"
            )



# """
#     redis_worker.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import json
import signal
from typing import Any, Optional

from snowplow_tracker.typing import EmitterProtocol, PayloadDict, RedisProtocol

_REDIS_OPT = True
try:
    import redis
    import gevent
    from gevent.pool import Pool
except ImportError:
    _REDIS_OPT = False

DEFAULT_KEY = "snowplow"


class RedisWorker(object):
    """
    Asynchronously take events from redis and send them to an emitter
    """

    if _REDIS_OPT:

        def __init__(
            self,
            emitter: EmitterProtocol,
            rdb: Optional[RedisProtocol] = None,
            key: str = DEFAULT_KEY,
        ) -> None:
            self.emitter = emitter
            self.key = key
            if rdb is None:
                rdb = redis.StrictRedis()
            self.rdb = rdb
            self.pool = Pool(5)

            signal.signal(signal.SIGTERM, self.request_shutdown)
            signal.signal(signal.SIGINT, self.request_shutdown)
            signal.signal(signal.SIGQUIT, self.request_shutdown)

        def send(self, payload: PayloadDict) -> None:
            """
            Send an event to an emitter
            """
            self.emitter.input(payload)

        def pop_payload(self) -> None:
            """
            Get a single event from Redis and send it
            If the Redis queue is empty, sleep to avoid making continual requests
            """
            payload = self.rdb.lpop(self.key)
            if payload:
                self.pool.spawn(self.send, json.loads(payload.decode("utf-8")))
            else:
                gevent.sleep(5)

        def run(self) -> None:
            """
            Run indefinitely
            """
            self._shutdown = False

            while not self._shutdown:
                self.pop_payload()
            self.pool.join(timeout=20)

        def request_shutdown(self, *args: Any) -> None:
            """
            Halt the worker
            """
            self._shutdown = True

    else:

        def __new__(cls, *args: Any, **kwargs: Any) -> "RedisWorker":
            raise RuntimeError(
                "RedisWorker is not available. To use: `pip install snowplow-tracker[redis]`"
            )



# """
#     _version.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

__version_info__ = (0, 13, 0)
__version__ = ".".join(str(x) for x in __version_info__)
__build_version__ = __version__ + ""



#
#     setup.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

authors_list = ["Anuj More", "Alexander Dean", "Fred Blundun", "Paul Boocock"]
authors_str = ", ".join(authors_list)

authors_email_list = [
    "support@snowplow.io",
]
authors_email_str = ", ".join(authors_email_list)

setup(
    name="snowplow-tracker",
    version="0.13.0",
    author=authors_str,
    author_email=authors_email_str,
    packages=[
        "snowplow_tracker",
        "snowplow_tracker.test",
        "snowplow_tracker.redis",
        "snowplow_tracker.celery",
    ],
    url="http://snowplow.io",
    license="Apache License 2.0",
    description="Snowplow event tracker for Python. Add analytics to your Python and Django apps, webapps and games",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    install_requires=["requests>=2.25.1,<3.0", "typing_extensions>=3.7.4"],
    extras_require={
        "celery": [
            "celery>=4.0,<5.0;python_version<'3.0'",
            "celery>=4.0;python_version>='3.0'",
        ],
        "redis": [
            "redis>=2.9.1,<4.0;python_version<'3.0'",
            "redis>=2.9.1;python_version>='3.0'",
            "gevent>=21.1.2",
        ],
    },
)



from snowplow_tracker._version import __version__
from snowplow_tracker.subject import Subject
from snowplow_tracker.emitters import logger, Emitter, AsyncEmitter
from snowplow_tracker.self_describing_json import SelfDescribingJson
from snowplow_tracker.tracker import Tracker
from snowplow_tracker.emitter_configuration import EmitterConfiguration
from snowplow_tracker.tracker_configuration import TrackerConfiguration
from snowplow_tracker.snowplow import Snowplow
from snowplow_tracker.contracts import disable_contracts, enable_contracts
from snowplow_tracker.event_store import EventStore

# celery extra
from .celery import CeleryEmitter

# redis extra
from .redis import RedisEmitter, RedisWorker



# """
#     celery_emitter.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

import logging
from typing import Any, Optional

from snowplow_tracker.emitters import Emitter
from snowplow_tracker.typing import HttpProtocol, Method

_CELERY_OPT = True
try:
    from celery import Celery
except ImportError:
    _CELERY_OPT = False

# logging
logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class CeleryEmitter(Emitter):
    """
    Uses a Celery worker to send HTTP requests asynchronously.
    Works like the base Emitter class,
    but on_success and on_failure callbacks cannot be set.
    """

    if _CELERY_OPT:

        celery_app = None

        def __init__(
            self,
            endpoint: str,
            protocol: HttpProtocol = "http",
            port: Optional[int] = None,
            method: Method = "post",
            batch_size: Optional[int] = None,
            byte_limit: Optional[int] = None,
        ) -> None:
            super(CeleryEmitter, self).__init__(
                endpoint, protocol, port, method, batch_size, None, None, byte_limit
            )

            try:
                # Check whether a custom Celery configuration module named "snowplow_celery_config" exists
                import snowplow_celery_config

                self.celery_app = Celery()
                self.celery_app.config_from_object(snowplow_celery_config)
            except ImportError:
                # Otherwise configure Celery with default settings
                self.celery_app = Celery("Snowplow", broker="redis://guest@localhost//")

            self.async_flush = self.celery_app.task(self.async_flush)

        def flush(self) -> None:
            """
            Schedules a flush task
            """
            self.async_flush.delay()
            logger.info("Scheduled a Celery task to flush the event queue")

        def async_flush(self) -> None:
            super(CeleryEmitter, self).flush()

    else:

        def __new__(cls, *args: Any, **kwargs: Any) -> "CeleryEmitter":
            logger.error(
                "CeleryEmitter is not available. Please install snowplow-tracker with celery extra dependency."
            )
            raise RuntimeError(
                "CeleryEmitter is not available. To use: `pip install snowplow-tracker[celery]`"
            )



# """
#     typing.py

#     Copyright (c) 2013-2023 Snowplow Analytics Ltd. All rights reserved.

#     This program is licensed to you under the Apache License Version 2.0,
#     and you may not use this file except in compliance with the Apache License
#     Version 2.0. You may obtain a copy of the Apache License Version 2.0 at
#     http://www.apache.org/licenses/LICENSE-2.0.

#     Unless required by applicable law or agreed to in writing,
#     software distributed under the Apache License Version 2.0 is distributed on
#     an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#     express or implied. See the Apache License Version 2.0 for the specific
#     language governing permissions and limitations there under.
# """

from typing import Dict, List, Callable, Any, Optional, Union, Tuple
from typing_extensions import Protocol, Literal

PayloadDict = Dict[str, Any]
PayloadDictList = List[PayloadDict]
JsonEncoderFunction = Callable[[Any], Any]

# tracker
FORM_NODE_NAMES = {"INPUT", "TEXTAREA", "SELECT"}
FORM_TYPES = {
    "button",
    "checkbox",
    "color",
    "date",
    "datetime",
    "datetime-local",
    "email",
    "file",
    "hidden",
    "image",
    "month",
    "number",
    "password",
    "radio",
    "range",
    "reset",
    "search",
    "submit",
    "tel",
    "text",
    "time",
    "url",
    "week",
}
FormNodeName = Literal["INPUT", "TEXTAREA", "SELECT"]
ElementClasses = Union[List[str], Tuple[str, Any]]
FormClasses = Union[List[str], Tuple[str, Any]]

# emitters
HttpProtocol = Literal["http", "https"]
Method = Literal["get", "post"]
SuccessCallback = Callable[[PayloadDictList], None]
FailureCallback = Callable[[int, PayloadDictList], None]

# subject
SUPPORTED_PLATFORMS = {"pc", "tv", "mob", "cnsl", "iot", "web", "srv", "app"}
SupportedPlatform = Literal["pc", "tv", "mob", "cnsl", "iot", "web", "srv", "app"]


class EmitterProtocol(Protocol):
    def input(self, payload: PayloadDict) -> None:
        ...


class RedisProtocol(Protocol):
    def rpush(self, name: Any, *values: Any) -> int:
        ...

    def lpop(self, name: Any, count: Optional[int] = ...) -> Any:
        ...
